#ifndef EDGE_COLREGS_H_
#define EDGE_COLREGS_H_

#define G2O_USE_VENDORED_CERES

#include "Vertex_Pose.hpp"
#include "Edge_TEB_Base.hpp"
#include "Penalty.hpp"
#include <g2o/core/base_unary_edge.h>


namespace teb_local_planner
{

//! Symbols for left/none/right rotations      
enum class RotType { left, none, right };

/**
 * @class EdgePreferRotDir
 * @brief Edge defining the cost function for penalzing a specified turning direction, in particular left resp. right turns
 * 
 * The edge depends on two consecutive vertices \f$ \mathbf{s}_i, \mathbf{s}_{i+1} \f$ and penalizes a given rotation direction
 * based on the \e weight and \e dir (\f$ dir \in \{-1,1\} \f$)
 * \e dir should be +1 to prefer left rotations and -1 to prefer right rotations  \n
 * \e weight can be set using setInformation(). \n
 * @see TebOptimalPlanner::AddEdgePreferRotDir
 */     
class EdgeColregs : public BaseTebBinaryEdge<1, double, VertexPose, VertexPose>
{
public:
    
  /**
   * @brief Construct edge.
   */    
  EdgeColregs() 
  {
    _measurement = 1;
  }
 
  /**
   * @brief Actual cost function
   */    
  void computeError()
  {
    const VertexPose* conf1 = static_cast<const VertexPose*>(_vertices[0]);
    const VertexPose* conf2 = static_cast<const VertexPose*>(_vertices[1]);
    
    _error[0] = penaltyBoundFromBelow( _measurement * g2o::normalize_theta(conf2->theta()-conf1->theta()) , 0, 0);
  }

  /**
   * @brief Specify the prefered direction of rotation
   * @param dir +1 to prefer the left side, -1 to prefer the right side
   */ 
  void setRotDir(double dir)
  {
    _measurement = dir;
  }
  
  /** Prefer rotations to the right */
  void preferRight() {_measurement = 1;}
    
  /** Prefer rotations to the right */
  void preferLeft() {_measurement = -1;}  
    
  
public: 
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW

};
  
} // end namespace

#endif