#ifndef _EDGE_KINEMATICS_H
#define _EDGE_KINEMATICS_H

#include "Vertex_Pose.hpp"
#include "Edge_TEB_Base.hpp"
#include "Penalty.hpp"
#include "../TEB_Config.hpp"

#include <cmath>

namespace teb_local_planner
{

/**
 * @class EdgeKinematicsDiffDrive
 * @brief Edge defining the cost function for satisfying the non-holonomic kinematics of a differential drive mobile robot.
 * 
 * The edge depends on two vertices \f$ \mathbf{s}_i, \mathbf{s}_{ip1} \f$ and minimizes a geometric interpretation
 * of the non-holonomic constraint: 
 * 	- C. Rösmann et al.: Trajectory modification considering dynamic constraints of autonomous robots, ROBOTIK, 2012.
 * 
 * The \e weight can be set using setInformation(): Matrix element 1,1: (Choose a very high value: ~1000). \n
 * A second equation is implemented to penalize backward motions (second element of the error /cost vector). \n
 * The \e weight can be set using setInformation(): Matrix element 2,2: (A value ~1 allows backward driving, but penalizes it slighly). \n
 * The dimension of the error / cost vector is 2: the first component represents the nonholonomic constraint cost, 
 * the second one backward-drive cost.
 * @see TebOptimalPlanner::AddEdgesKinematics, EdgeKinematicsCarlike
 * @remarks Do not forget to call setTebConfig()
 */    
class EdgeKinematics : public BaseTebBinaryEdge<2, double, VertexPose, VertexPose>
{
public:
  
  /**
   * @brief Construct edge.
   */  
  EdgeKinematics()
  {
      this->setMeasurement(0.);
  }
  
  /**
   * @brief Actual cost function
   */    
  void computeError()
  {
    const VertexPose* conf1 = static_cast<const VertexPose*>(_vertices[0]);
    const VertexPose* conf2 = static_cast<const VertexPose*>(_vertices[1]);
    
    Eigen::Vector2d deltaS = conf2->position() - conf1->position();

    // non holonomic constraint
    _error[0] = fabs( ( cos(conf1->theta())+cos(conf2->theta()) ) * deltaS[1] - ( sin(conf1->theta())+sin(conf2->theta()) ) * deltaS[0] );

    // // positive-drive-direction constraint
    Eigen::Vector2d angle_vec ( cos(conf1->theta()), sin(conf1->theta()) );
    _error[1] = penaltyBoundFromBelow(deltaS.dot(angle_vec), 0,0);
    // epsilon=0, otherwise it pushes the first bandpoints away from start
  }

#ifdef USE_ANALYTIC_JACOBI
// #if 0
  /**
   * @brief Jacobi matrix of the cost function specified in computeError().
   */
  void linearizeOplus()
  {
    const VertexPose* conf1 = static_cast<const VertexPose*>(_vertices[0]);
    const VertexPose* conf2 = static_cast<const VertexPose*>(_vertices[1]);
    
    Eigen::Vector2d deltaS = conf2->position() - conf1->position();
	    
    double cos1 = cos(conf1->theta());
    double cos2 = cos(conf2->theta());
    double sin1 = sin(conf1->theta());
    double sin2 = sin(conf2->theta());
    double aux1 = sin1 + sin2;
    double aux2 = cos1 + cos2;
    
    double dd_error_1 = deltaS[0]*cos1;
    double dd_error_2 = deltaS[1]*sin1;
    double dd_dev = penaltyBoundFromBelowDerivative(dd_error_1+dd_error_2, 0,0);
    
    double dev_nh_abs = g2o::sign( ( cos(conf1->theta())+cos(conf2->theta()) ) * deltaS[1] - 
	      ( sin(conf1->theta())+sin(conf2->theta()) ) * deltaS[0] );
	    
    // conf1
    _jacobianOplusXi(0,0) = aux1 * dev_nh_abs; // nh x1
    _jacobianOplusXi(0,1) = -aux2 * dev_nh_abs; // nh y1
    _jacobianOplusXi(1,0) = -cos1 * dd_dev; // drive-dir x1
    _jacobianOplusXi(1,1) = -sin1 * dd_dev; // drive-dir y1
    _jacobianOplusXi(0,2) = (-dd_error_2 - dd_error_1) * dev_nh_abs; // nh angle
    _jacobianOplusXi(1,2) = ( -sin1*deltaS[0] + cos1*deltaS[1] ) * dd_dev; // drive-dir angle1
    
    // conf2
    _jacobianOplusXj(0,0) = -aux1 * dev_nh_abs; // nh x2
    _jacobianOplusXj(0,1) = aux2 * dev_nh_abs; // nh y2
    _jacobianOplusXj(1,0) = cos1 * dd_dev; // drive-dir x2
    _jacobianOplusXj(1,1) = sin1 * dd_dev; // drive-dir y2
    _jacobianOplusXj(0,2) = (-sin2*deltaS[1] - cos2*deltaS[0]) * dev_nh_abs; // nh angle
    _jacobianOplusXj(1,2) = 0; // drive-dir angle1					
  }
// #endif
#endif
      
public:
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW   
};

} // end namespace

#endif