#ifndef EDGE_OBSTACLE_H_
#define EDGE_OBSTACLE_H_

#include "Vertex_Pose.hpp"
#include "Edge_TEB_Base.hpp"
#include "Penalty.hpp"
#include "../TEB_Config.hpp"

namespace teb_local_planner
{

/**
 * @class EdgeObstacle
 * @brief Edge defining the cost function for keeping a minimum distance from obstacles.
 * 
 * The edge depends on a single vertex \f$ \mathbf{s}_i \f$ and minimizes: \n
 * \f$ \min \textrm{penaltyBelow}( dist2point ) \cdot weight \f$. \n
 * \e dist2point denotes the minimum distance to the point obstacle. \n
 * \e weight can be set using setInformation(). \n
 * \e penaltyBelow denotes the penalty function, see penaltyBoundFromBelow() \n
 * @see TebOptimalPlanner::AddEdgesObstacles, TebOptimalPlanner::EdgeInflatedObstacle
 * @remarks Do not forget to call setTebConfig() and setObstacle()
 */     
class EdgeObstacle : public BaseTebBinaryEdge<1, const std::vector<Eigen::Vector2d>*, VertexPose, VertexPose>
{
public:
    
  /**
   * @brief Construct edge.
   */    
  EdgeObstacle() 
  {
    _measurement = NULL;
  }
 
  /**
   * @brief Actual cost function
   */    
  void computeError()
  {
    const VertexPose* bandpt1 = static_cast<const VertexPose*>(_vertices[0]);
    const VertexPose* bandpt2 = static_cast<const VertexPose*>(_vertices[1]);

    auto delta = 1/cfg_->obstacles.num * (bandpt2->position() - bandpt1->position());

    if (_measurement->size() == 0)
    {
      _error[0] = 0;
    }
    else
    {
      _error[0] = 0;
      for (int i = 0; i < cfg_->obstacles.num+1; ++i)
      {
        auto position = bandpt1->position() + i * delta;
        for (int j=0; j<_measurement->size(); ++j)
        {
          double dist = (position - _measurement->at(j)).norm();
           _error[0] += penaltyBoundFromBelow(dist, cfg_->obstacles.min_obstacle_dist, cfg_->optim.penalty_epsilon);
        }
      }
    }
  }

  /**
   * @brief Set all parameters at once
   * @param cfg TebConfig class
   * @param robot_model Robot model required for distance calculation
   * @param obstacle 2D position vector containing the position of the obstacle
   */ 
  void setParameters(const TebConfig& cfg, const std::vector<Eigen::Vector2d>& obstacle)
  {
    cfg_ = &cfg;
    _measurement = &obstacle;
  }
  
public: 	
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW

};

}

#endif

// namespace teb_local_planner
// {

// /**
//  * @class EdgeObstacle
//  * @brief Edge defining the cost function for keeping a minimum distance from obstacles.
//  * 
//  * The edge depends on a single vertex \f$ \mathbf{s}_i \f$ and minimizes: \n
//  * \f$ \min \textrm{penaltyBelow}( dist2point ) \cdot weight \f$. \n
//  * \e dist2point denotes the minimum distance to the point obstacle. \n
//  * \e weight can be set using setInformation(). \n
//  * \e penaltyBelow denotes the penalty function, see penaltyBoundFromBelow() \n
//  * @see TebOptimalPlanner::AddEdgesObstacles, TebOptimalPlanner::EdgeInflatedObstacle
//  * @remarks Do not forget to call setTebConfig() and setObstacle()
//  */     
// class EdgeObstacle : public BaseTebUnaryEdge<1, const std::vector<Eigen::Vector2d>*, VertexPose>
// {
// public:
    
//   /**
//    * @brief Construct edge.
//    */    
//   EdgeObstacle() 
//   {
//     _measurement = NULL;
//   }
 
//   /**
//    * @brief Actual cost function
//    */    
//   void computeError()
//   {
//     const VertexPose* bandpt = static_cast<const VertexPose*>(_vertices[0]);

//     if (_measurement->size() == 0)
//     {
//       _error[0] = 0;
//     }
//     else
//     {
//       double dist_min = getMinDistance(*_measurement, bandpt->position()).first;

//       // Original obstacle cost.
//       _error[0] = penaltyBoundFromBelow(dist_min, cfg_->obstacles.min_obstacle_dist, cfg_->optim.penalty_epsilon);
//     }
//   }

//   /**
//    * @brief Set all parameters at once
//    * @param cfg TebConfig class
//    * @param robot_model Robot model required for distance calculation
//    * @param obstacle 2D position vector containing the position of the obstacle
//    */ 
//   void setParameters(const TebConfig& cfg, const std::vector<Eigen::Vector2d>& obstacle)
//   {
//     cfg_ = &cfg;
//     _measurement = &obstacle;
//   }
  
// public: 	
//   EIGEN_MAKE_ALIGNED_OPERATOR_NEW

// };

// }

// #endif