#ifndef EDGE_TIMEOPTIMAL_H_
#define EDGE_TIMEOPTIMAL_H_

#include <float.h>

#include "Vertex_Time.hpp"
#include "Edge_TEB_Base.hpp"
#include "Penalty.hpp"
#include "../TEB_Config.hpp"

#include <Eigen/Core>

namespace teb_local_planner
{

/**
 * @class EdgeTimeOptimal
 * @brief Edge defining the cost function for minimizing transition time of the trajectory.
 * 
 * The edge depends on a single vertex \f$ \Delta T_i \f$ and minimizes: \n
 * \f$ \min \Delta T_i^2 \cdot scale \cdot weight \f$. \n
 * \e scale is determined using the penaltyEquality() function, since we experiences good convergence speeds with it. \n
 * \e weight can be set using setInformation() (something around 1.0 seems to be fine). \n
 * @see TebOptimalPlanner::AddEdgesTimeOptimal
 * @remarks Do not forget to call setTebConfig()
 */
class EdgeTimeOptimal : public BaseTebUnaryEdge<1, double, VertexTimeDiff>
{
public:
    
  /**
   * @brief Construct edge.
   */
  EdgeTimeOptimal()
  {
    this->setMeasurement(0.);
  }
  
  /**
   * @brief Actual cost function
   */
  void computeError()
  {
    const VertexTimeDiff* timediff = static_cast<const VertexTimeDiff*>(_vertices[0]);

   _error[0] = timediff->dt();
  
  }

#ifdef USE_ANALYTIC_JACOBI
  /**
   * @brief Jacobi matrix of the cost function specified in computeError().
   */
  void linearizeOplus()
  {
    _jacobianOplusXi( 0 , 0 ) = 1;
  }
#endif
  
  
public:        
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW
};

}; // end namespace

#endif /* EDGE_TIMEOPTIMAL_H_ */