#include "../include/Optimal_Planner.hpp"

#include <memory>
#include <limits>
#include <boost/thread/thread.hpp>
#include <boost/thread/once.hpp>

namespace teb_local_planner
{

// ============== Implementation ===================

TebOptimalPlanner::TebOptimalPlanner() {}

TebOptimalPlanner::TebOptimalPlanner(TebConfig* cfg)
{    
  optimizer_ = initOptimizer();
  cfg_ = cfg;
  cost_ = HUGE_VAL;
  initialized_ = true;
}

TebOptimalPlanner::~TebOptimalPlanner()
{
  clearGraph();
}

/*
 * registers custom vertices and edges in g2o framework
 */
void TebOptimalPlanner::registerG2OTypes()
{
  g2o::Factory* factory = g2o::Factory::instance();
  factory->registerType("VERTEX_POSE", std::make_shared<g2o::HyperGraphElementCreator<VertexPose>>());
  factory->registerType("VERTEX_TIMEDIFF", std::make_shared<g2o::HyperGraphElementCreator<VertexTimeDiff>>());

  factory->registerType("EDGE_TIME_OPTIMAL", std::make_shared<g2o::HyperGraphElementCreator<EdgeTimeOptimal>>());
  factory->registerType("EDGE_VELOCITY", std::make_shared<g2o::HyperGraphElementCreator<EdgeVelocity>>());
  factory->registerType("EDGE_ACCELERATION", std::make_shared<g2o::HyperGraphElementCreator<EdgeAcceleration>>());
  factory->registerType("EDGE_ACCELERATION_START", std::make_shared<g2o::HyperGraphElementCreator<EdgeAccelerationStart>>());
  factory->registerType("EDGE_ACCELERATION_GOAL", std::make_shared<g2o::HyperGraphElementCreator<EdgeAccelerationGoal>>());
  factory->registerType("EDGE_KINEMATICS", std::make_shared<g2o::HyperGraphElementCreator<EdgeKinematics>>());
  factory->registerType("EDGE_OBSTACLE", std::make_shared<g2o::HyperGraphElementCreator<EdgeObstacle>>());
  factory->registerType("EDGE_COLREGS", std::make_shared<g2o::HyperGraphElementCreator<EdgeColregs>>());
  factory->registerType("EDGE_TURNING_RADIUS", std::make_shared<g2o::HyperGraphElementCreator<EdgeTurningRadius>>());
  factory->registerType("EDGE_GOAL", std::make_shared<g2o::HyperGraphElementCreator<EdgeGoal>>());
}

/*
 * initialize g2o optimizer. Set solver settings here.
 * Return: pointer to new SparseOptimizer Object.
 */
boost::shared_ptr<g2o::SparseOptimizer> TebOptimalPlanner::initOptimizer()
{
  // Call register_g2o_types once, even for multiple TebOptimalPlanner instances (thread-safe)
  static boost::once_flag flag = BOOST_ONCE_INIT;
  boost::call_once(&registerG2OTypes, flag);  

  // allocating the optimizer
  boost::shared_ptr<g2o::SparseOptimizer> optimizer = boost::make_shared<g2o::SparseOptimizer>();
  std::unique_ptr<TEBLinearSolver> linear_solver(new TEBLinearSolver());
  linear_solver->setBlockOrdering(true);
  std::unique_ptr<TEBBlockSolver> block_solver(new TEBBlockSolver(std::move(linear_solver)));
  g2o::OptimizationAlgorithmLevenberg* solver = new g2o::OptimizationAlgorithmLevenberg(std::move(block_solver));

  optimizer->setAlgorithm(solver);
  
  optimizer->initMultiThreading(); // required for >Eigen 3.1
  
  return optimizer;
}

void TebOptimalPlanner::initPlanner(PoseSE2& pose_start, PoseSE2& pose_goal, Eigen::Vector3d& vel_start, Eigen::Vector3d& vel_goal, std::vector<Eigen::Vector2d>& obstacles, RotType& prefer_rotdir)
{
  pose_start_ =  pose_start;
  pose_goal_ = pose_goal;

  vel_start_ = vel_start;
  vel_goal_ = vel_goal;

  obstacles_ = obstacles;

  prefer_rotdir_ = prefer_rotdir;
}

bool TebOptimalPlanner::optimizeTEB(int iterations_innerloop, int iterations_outerloop, bool compute_cost_afterwards,
                                    double obst_cost_scale, bool alternative_time_cost)
{
  if (cfg_->optim.optimization_activate==false) 
    return false;
  
  bool success = false;
  optimized_ = false;
  bool fast_mode = false;
  
  for(int i=0; i<iterations_outerloop; ++i)
  {
    if (cfg_->trajectory.teb_autosize)
    {
      teb_.autoResize(cfg_->trajectory.dt_ref, cfg_->trajectory.dt_hysteresis, cfg_->trajectory.min_samples, cfg_->trajectory.max_samples, fast_mode);
    }

    success = buildGraph();
    if (!success) 
    {
        clearGraph();
        return false;
    }
    success = optimizeGraph(iterations_innerloop, false);
    if (!success) 
    {
        clearGraph();
        return false;
    }
    optimized_ = true;
      
    clearGraph();
  }

  return true;
}


bool TebOptimalPlanner::plan(PoseSE2& pose_start, PoseSE2& pose_goal, Eigen::Vector3d& vel_start, Eigen::Vector3d& vel_goal, std::vector<Eigen::Vector2d>& obstacles, RotType& prefer_rotdir)
{	
  initPlanner(pose_start, pose_goal, vel_start, vel_goal, obstacles, prefer_rotdir);

  if (!teb_.isInit())
  {
    //init trajectory
    teb_.initTrajectoryToGoal(pose_start_, pose_goal_, cfg_->trajectory.distance_step,
                              cfg_->robot.max_vel_x, cfg_->trajectory.min_samples);
  }
  else // warm start
  {
    if (teb_.sizePoses() > 0
        && (pose_goal_.position() - teb_.BackPose().position()).norm() < cfg_->trajectory.force_reinit_new_goal_dist
        && fabs(g2o::normalize_theta(pose_goal_.theta() - teb_.BackPose().theta())) < cfg_->trajectory.force_reinit_new_goal_angular)
        teb_.updateAndPruneTEB(pose_start_, pose_goal_, cfg_->trajectory.min_samples);
    else
    {
        teb_.clearTimedElasticBand();
        teb_.initTrajectoryToGoal(pose_start_, pose_goal_, cfg_->trajectory.distance_step,
                              cfg_->robot.max_vel_x, cfg_->trajectory.min_samples);
    }
  }

  return optimizeTEB(cfg_->optim.no_inner_iterations, cfg_->optim.no_outer_iterations);
}


bool TebOptimalPlanner::buildGraph()
{
  if (!optimizer_->edges().empty() || !optimizer_->vertices().empty())
  {
    return false;
  }
  
  // add vertices
  AddTEBVertices();
  
  // add Edges
  AddEdgesObstacles();
  
  AddEdgesVelocity();
  
  AddEdgesAcceleration();

  // AddEdgesTimeOptimal();
  
  AddEdgesKinematics();

  AddEdgesColregs();

  AddEdgesTurningRadius();
    
  AddEdgeGoal();

  return true;  
}

bool TebOptimalPlanner::optimizeGraph(int no_iterations,bool clear_after)
{
  if (cfg_->robot.max_vel_x<0.01)
  {
    if (clear_after) clearGraph();
    return false;	
  }
  
  if (!teb_.isInit() || teb_.sizePoses() < cfg_->trajectory.min_samples)
  {
    if (clear_after) clearGraph();
    return false;	
  }
  
  optimizer_->setVerbose(cfg_->optim.optimization_verbose);
  optimizer_->initializeOptimization();

  int iter = optimizer_->optimize(no_iterations);

  if(!iter)
  {
	return false;
  }

  if (clear_after) clearGraph();	
    
  return true;
}

void TebOptimalPlanner::clearGraph()
{
  // clear optimizer states
  if (optimizer_)
  {
    // we will delete all edges but keep the vertices.
    // before doing so, we will delete the link from the vertices to the edges.
    auto& vertices = optimizer_->vertices();
    for(auto& v : vertices)
      v.second->edges().clear();

    // necessary, because optimizer->clear deletes pointer-targets (therefore it deletes TEB states!)
    optimizer_->vertices().clear();  
    optimizer_->clear();
  }
}

void TebOptimalPlanner::AddTEBVertices()
{
  // add vertices to graph
  unsigned int id_counter = 0; // used for vertices ids
  for (int i=0; i<teb_.sizePoses(); ++i)
  {
    teb_.PoseVertex(i)->setId(id_counter++);
    optimizer_->addVertex(teb_.PoseVertex(i));
    if (teb_.sizeTimeDiffs()!=0 && i<teb_.sizeTimeDiffs())
    {
      teb_.TimeDiffVertex(i)->setId(id_counter++);
      optimizer_->addVertex(teb_.TimeDiffVertex(i));
    }
  }
}


void TebOptimalPlanner::AddEdgesObstacles()
{
  if (cfg_->optim.weight_obstacle==0)
    return; // if weight equals zero skip adding edges!

  int n = teb_.sizePoses();
  Eigen::Matrix<double,1,1> information;
  information.fill(cfg_->optim.weight_obstacle);
  for (int i=0; i < n-1; ++i)
  {
    EdgeObstacle* dist_bandpt_obst = new EdgeObstacle;
    dist_bandpt_obst->setVertex(0,teb_.PoseVertex(i));
    dist_bandpt_obst->setVertex(1,teb_.PoseVertex(i+1));
    dist_bandpt_obst->setInformation(information);
    dist_bandpt_obst->setParameters(*cfg_, obstacles_);
    optimizer_->addEdge(dist_bandpt_obst);
  };
}

void TebOptimalPlanner::AddEdgesVelocity()
{

  if ( cfg_->optim.weight_max_vel_x==0 && cfg_->optim.weight_max_vel_theta==0)
    return; // if weight equals zero skip adding edges!

  int n = teb_.sizePoses();
  Eigen::Matrix<double,2,2> information;
  information(0,0) = cfg_->optim.weight_max_vel_x;
  information(1,1) = cfg_->optim.weight_max_vel_theta;
  information(0,1) = 0.0;
  information(1,0) = 0.0;

  for (int i=0; i < n - 1; ++i)
  {
    EdgeVelocity* velocity_edge = new EdgeVelocity;
    velocity_edge->setVertex(0,teb_.PoseVertex(i));
    velocity_edge->setVertex(1,teb_.PoseVertex(i+1));
    velocity_edge->setVertex(2,teb_.TimeDiffVertex(i));
    velocity_edge->setInformation(information);
    velocity_edge->setTebConfig(*cfg_);
    optimizer_->addEdge(velocity_edge);
  }

}

void TebOptimalPlanner::AddEdgesAcceleration()
{
  if (cfg_->optim.weight_acc_lim_x==0  && cfg_->optim.weight_acc_lim_theta==0) 
    return; // if weight equals zero skip adding edges!

  int n = teb_.sizePoses();  
    
  Eigen::Matrix<double,2,2> information;
  information.fill(0);
  information(0,0) = cfg_->optim.weight_acc_lim_x;
  information(1,1) = cfg_->optim.weight_acc_lim_theta;
  
  EdgeAccelerationStart* acceleration_start_edge = new EdgeAccelerationStart;
  acceleration_start_edge->setVertex(0,teb_.PoseVertex(0));
  acceleration_start_edge->setVertex(1,teb_.PoseVertex(1));
  acceleration_start_edge->setVertex(2,teb_.TimeDiffVertex(0));
  acceleration_start_edge->setInitialVelocity(vel_start_);
  acceleration_start_edge->setInformation(information);
  acceleration_start_edge->setTebConfig(*cfg_);
  optimizer_->addEdge(acceleration_start_edge);

  // now add the usual acceleration edge for each tuple of three teb poses
  for (int i=0; i < n - 2; ++i)
  {
    EdgeAcceleration* acceleration_edge = new EdgeAcceleration;
    acceleration_edge->setVertex(0,teb_.PoseVertex(i));
    acceleration_edge->setVertex(1,teb_.PoseVertex(i+1));
    acceleration_edge->setVertex(2,teb_.PoseVertex(i+2));
    acceleration_edge->setVertex(3,teb_.TimeDiffVertex(i));
    acceleration_edge->setVertex(4,teb_.TimeDiffVertex(i+1));
    acceleration_edge->setInformation(information);
    acceleration_edge->setTebConfig(*cfg_);
    optimizer_->addEdge(acceleration_edge);
  }
    
  EdgeAccelerationGoal* acceleration_goal_edge = new EdgeAccelerationGoal;
  acceleration_goal_edge->setVertex(0,teb_.PoseVertex(n-2));
  acceleration_goal_edge->setVertex(1,teb_.PoseVertex(n-1));
  acceleration_goal_edge->setVertex(2,teb_.TimeDiffVertex(teb_.sizeTimeDiffs()-1));
  acceleration_goal_edge->setGoalVelocity(vel_goal_);
  acceleration_goal_edge->setInformation(information);
  acceleration_goal_edge->setTebConfig(*cfg_);
  optimizer_->addEdge(acceleration_goal_edge);
}

void TebOptimalPlanner::AddEdgesTimeOptimal()
{
  if (cfg_->optim.weight_optimaltime==0) 
    return; // if weight equals zero skip adding edges!

  Eigen::Matrix<double,1,1> information;
  information.fill(cfg_->optim.weight_optimaltime);

  for (int i=0; i < teb_.sizeTimeDiffs(); ++i)
  {
    EdgeTimeOptimal* timeoptimal_edge = new EdgeTimeOptimal;
    timeoptimal_edge->setVertex(0,teb_.TimeDiffVertex(i));
    timeoptimal_edge->setInformation(information);
    timeoptimal_edge->setTebConfig(*cfg_);
    optimizer_->addEdge(timeoptimal_edge);
  }
}


void TebOptimalPlanner::AddEdgesKinematics()
{
  if (cfg_->optim.weight_kinematics_nh==0 && cfg_->optim.weight_kinematics_forward_drive==0)
    return; // if weight equals zero skip adding edges!
  
  // create edge for satisfiying kinematic constraints
  Eigen::Matrix<double,2,2> information_kinematics;
  information_kinematics.fill(0.0);
  information_kinematics(0, 0) = cfg_->optim.weight_kinematics_nh;
  information_kinematics(1, 1) = cfg_->optim.weight_kinematics_forward_drive;
  
  for (int i=0; i < teb_.sizePoses()-1; i++) // ignore twiced start only
  {
    EdgeKinematics* kinematics_edge = new EdgeKinematics;
    kinematics_edge->setVertex(0,teb_.PoseVertex(i));
    kinematics_edge->setVertex(1,teb_.PoseVertex(i+1));      
    kinematics_edge->setInformation(information_kinematics);
    kinematics_edge->setTebConfig(*cfg_);
    optimizer_->addEdge(kinematics_edge);
  }	 
}

void TebOptimalPlanner::AddEdgesColregs()
{
  if (prefer_rotdir_ == RotType::none || cfg_->optim.weight_colregs==0)
    return; // if weight equals zero skip adding edges!

  if (prefer_rotdir_ != RotType::right && prefer_rotdir_ != RotType::left)
  {
    return;
  }

  // create edge for satisfiying kinematic constraints
  Eigen::Matrix<double,1,1> information_colregs;
  information_colregs.fill(cfg_->optim.weight_colregs);
  
  for (int i=0; i < teb_.sizePoses()-1 && i < 2; ++i) // currently: apply to first 5 rotations
  {
    EdgeColregs* rotdir_edge = new EdgeColregs;
    rotdir_edge->setVertex(0,teb_.PoseVertex(i));
    rotdir_edge->setVertex(1,teb_.PoseVertex(i+1));      
    rotdir_edge->setInformation(information_colregs);
    
    if (prefer_rotdir_ == RotType::left)
        rotdir_edge->preferLeft();
    else if (prefer_rotdir_ == RotType::right)
        rotdir_edge->preferRight();
    
    optimizer_->addEdge(rotdir_edge);
  }
}

void TebOptimalPlanner::AddEdgesTurningRadius()
{
    if ( cfg_->optim.weight_min_turning_radius==0)
    return; // if weight equals zero skip adding edges!

  int n = teb_.sizePoses();
  Eigen::Matrix<double,1,1> information_turningRadius;
  information_turningRadius.fill(cfg_->optim.weight_min_turning_radius);

  for (int i=0; i < n - 1; ++i)
  {
    EdgeTurningRadius* turningRadius_edge = new EdgeTurningRadius;
    turningRadius_edge->setVertex(0,teb_.PoseVertex(i));
    turningRadius_edge->setVertex(1,teb_.PoseVertex(i+1));
    turningRadius_edge->setVertex(2,teb_.TimeDiffVertex(i));
    turningRadius_edge->setInformation(information_turningRadius);
    turningRadius_edge->setTebConfig(*cfg_);
    optimizer_->addEdge(turningRadius_edge);
  }
}

void TebOptimalPlanner::AddEdgeGoal()
{
    if ( cfg_->optim.weight_goal==0)
    return; // if weight equals zero skip adding edges!

  int n = teb_.sizePoses();
  Eigen::Matrix<double,1,1> information_goal;
  information_goal.fill(cfg_->optim.weight_goal);

  EdgeGoal* goal_edge = new EdgeGoal;
  goal_edge->setVertex(0,teb_.PoseVertex(n-1));
  goal_edge->setInformation(information_goal);
  goal_edge->setParameter(*cfg_,pose_goal_);
  optimizer_->addEdge(goal_edge);
}

void TebOptimalPlanner::extractVelocity(const PoseSE2& pose1, const PoseSE2& pose2, double dt, double& vx, double& vy, double& omega) const
{
  if (dt == 0)
  {
    vx = 0;
    vy = 0;
    omega = 0;
    return;
  }

  Eigen::Vector2d deltaS = pose2.position() - pose1.position();
  double orientdiff = g2o::normalize_theta(pose2.theta() - pose1.theta());
  double dist = deltaS.norm();
  if (orientdiff != 0)
  {
      double radius =  dist/(2*sin(orientdiff/2));
      dist = fabs( orientdiff * radius ); // actual arg length!
  }
  
  vx = dist/dt;
  vy = 0;
  omega = orientdiff/dt;
}


Trajectory TebOptimalPlanner::getTrajectory() const
{
  Trajectory trajectory;
  double curr_time = 0;
  int n = teb_.sizePoses();

  std::tuple<double, PoseSE2, Eigen::Vector3d> start_point = std::make_tuple(curr_time,teb_.Pose(0),vel_start_);
  trajectory.push_back(start_point);
  curr_time += teb_.TimeDiff(0);

  // intermediate points
  for (int i=1; i < n-1; ++i)
  {
    double vel1_x, vel1_y, vel2_x, vel2_y, omega1, omega2;

    extractVelocity(teb_.Pose(i-1), teb_.Pose(i), teb_.TimeDiff(i-1), vel1_x, vel1_y, omega1);
    extractVelocity(teb_.Pose(i), teb_.Pose(i+1), teb_.TimeDiff(i), vel2_x, vel2_y, omega2);
    Eigen::Vector3d velocity((vel1_x+vel2_x)/2, (vel1_y+vel2_y)/2, (omega1+omega2)/2);

    std::tuple<double, PoseSE2, Eigen::Vector3d> inter_point = std::make_tuple(curr_time,teb_.Pose(i),velocity);
    trajectory.push_back(inter_point);
    curr_time += teb_.TimeDiff(i);
  }
  
  // goal
  std::tuple<double, PoseSE2, Eigen::Vector3d> goal_point = std::make_tuple(curr_time,teb_.BackPose(),vel_goal_);
  trajectory.push_back(goal_point);

  return trajectory;
}


Eigen::VectorXd TebOptimalPlanner::getCommand(Trajectory& trajectory) const
{
  int idx = 0;

  while (true)
  {
    if (cfg_->planning.execution_time > std::get<0>(trajectory[idx]))
      idx++;
    else
    {
      idx--;
      break;
    }
  }

  double delta_T =  std::get<0>(trajectory[idx+1]) - std::get<0>(trajectory[idx]);
  double r_dot = (std::get<2>(trajectory[idx+1])(2) - std::get<2>(trajectory[idx])(2)) / delta_T;
  double r_c = std::get<2>(trajectory[idx])(2) + r_dot * (cfg_->planning.execution_time - std::get<0>(trajectory[idx]));
  double psi_c = g2o::normalize_theta(std::get<1>(trajectory[idx]).theta() + (pow(r_c,2)-pow(std::get<2>(trajectory[idx])(2),2) )/(2*r_dot));
  double r_dot_c = (r_c - std::get<2>(trajectory.front())[2]) / cfg_->planning.execution_time;

  double u_dot = (std::get<2>(trajectory[idx+1])(0) - std::get<2>(trajectory[idx])(0)) / delta_T;
  double u_c = std::get<2>(trajectory[idx])(0) + u_dot * (cfg_->planning.execution_time - std::get<0>(trajectory[idx]));
  double u_dot_c = (u_c - std::get<2>(trajectory.front())[0]) / cfg_->planning.execution_time;

  Eigen::VectorXd command(5);
  command << psi_c, r_c, r_dot_c, u_c, u_dot_c;

  return command;
}

}