#include "../include/Ship.hpp"

Ship::Ship(ShipConfig& ship_cfg):
    usv(0,0), guidance(ship_cfg.waypoints), eta_6(6), nu_6(6)
{
    ship_idx = ship_cfg.index;

    eta_3 = ship_cfg.eta0;
    nu_3 = ship_cfg.nu0;
    eta_6 << eta_3.x(), eta_3.y(), 0, 0, 0, eta_3.theta();
    nu_6 << nu_3(0), nu_3(1), 0, 0, 0, nu_3(2);

    psi_ref = eta_3.theta();
    psi_d = eta_3.theta();
    r_d = nu_3(2);
    alpha_d = 0;

    u_nominal = ship_cfg.u_nominal;
    u_ref = ship_cfg.u_nominal;
    u_d = nu_3(0);
    a_d = 0;

    n << 0, 0;
    n_c << 0, 0;

    wps = ship_cfg.waypoints;
    map = ship_cfg.map;

    safe = true;
}

void Ship::recordData()
{
    record(eta_3.x(), data_x);
    record(eta_3.y(), data_y);
    record(R2D(eta_3.theta()), data_psi);
    record(R2D(psi_ref), data_psi_ref);
    record(R2D(psi_d), data_psi_d);
    record(nu_3(0),data_u);
    record(u_ref, data_u_ref);
    record(u_d, data_u_d);
}

void Ship::recordPose()
{
    record(eta_3.x(), data_x_);
    record(eta_3.y(), data_y_);
    record(eta_3.theta(), data_psi_);
}

std::pair<PoseSE2,Eigen::Vector3d> Ship::provideMotionData()
{
    return {eta_3, nu_3};
}


std::string Ship::getMode()
{   
    COLREGs colregs(motion_data,ship_idx);
    colregs.updateState(safe);
    bool stand_on;
    prefer_rotdir = RotType::none;
    encounter_types.clear();

    if ( !safe )
    {
        colregs.updateEncounterType();
        encounter_types = colregs.encounter_types;
        auto action = colregs.getAction();
        stand_on = action.first;
        prefer_rotdir = action.second;

        if (!stand_on)
            return "Collision Avoidance";
    }


    for (int i=0; i<map.size(); ++i)
    {
        double distance = (eta_3.position()-map[i]).norm();
        if ( distance < 2*cfg.obstacles.min_obstacle_dist )
        {
            safe = false;
            if (!stand_on || distance < 0.5*cfg.obstacles.min_obstacle_dist)
                return "Collision Avoidance";
        }
    }

    return "Path Following";
}


void Ship::collisionAvoidance()
{   
    if (guidance.reached_goal)
    {
        pathFollowing();
        return;
    }

    if ( int(100*time)%100 == 0) 
    {   
        clock_t start, end;
        start = clock();
        
        MotionPrediction motion_predictor(motion_data,ship_idx);
        TebOptimalPlanner TEBplanner(&cfg);

        auto dynamic_obstacles = motion_predictor.predict();
        auto range = guidance.goal_dist + cfg.obstacles.min_obstacle_dist;
        auto obstacles = getNearbyObstacles(eta_3.position(), map, dynamic_obstacles, range);
        auto [eta_goal, nu_goal] = guidance.getLocalGoal(eta_3,u_ref);
        
        TEBplanner.plan(eta_3, eta_goal, nu_3, nu_goal, obstacles, prefer_rotdir);
        auto trajectory = TEBplanner.getTrajectory();
        auto command = TEBplanner.getCommand(trajectory);

        psi_d = command(0);
        r_d = command(1);
        alpha_d = command(2);
        u_d = sat(command(3),cfg.robot.min_vel_x,cfg.robot.max_vel_x);
        a_d = command(4);

        psi_ref = psi_d;
        u_ref = u_d;

        end = clock();
        double time_taken = double(end - start) / double(CLOCKS_PER_SEC);

        for (int i = 0; i < encounter_types.size(); ++i){
            std::cout << "ASV " + std::to_string(ship_idx+1) << ": " << encounter_types[i] << std::endl;
        }
        if (prefer_rotdir == RotType::right){
            std::cout << "ASV " + std::to_string(ship_idx+1) << ": Right turn" << std::endl;
        }

        std::cout << std::setprecision(3);
        std::cout << "ASV " + std::to_string(ship_idx+1) << ": time = " << time<< " psi_ref = " << R2D(psi_d) << " u_ref = " << u_d << std::endl;
        std::cout << "Runtime: " << std::fixed << time_taken << "s" << std::endl << "\n";
    }
}


void Ship::pathFollowing()
{
    if (guidance.reached_goal)
        u_nominal = 0;
    
    psi_ref = guidance.LOS(eta_3);
    guidance.referenceModel(psi_ref, psi_d, r_d, alpha_d, u_nominal, u_d, a_d);
    u_ref = u_nominal;
}


void Ship::simulate(double& current_time, std::vector<std::pair<PoseSE2,Eigen::Vector3d>>& ship_motion_data)
{   
    time = current_time;
    motion_data = ship_motion_data;

    recordData();
    if ( int(100*time) % 2000 == 0)
        recordPose();
    auto mode = getMode();
    if (mode == "Path Following")
        pathFollowing();
    else if (mode == "Collision Avoidance")
        collisionAvoidance();
    
    usv.simulate(eta_6, nu_6, n, n_c, psi_d, u_d, r_d, a_d);

    eta_3 = PoseSE2(eta_6(0), eta_6(1), eta_6(5));
    nu_3 = Eigen::Vector3d(nu_6(0), nu_6(1), nu_6(5));
}


void Ship::simulate(double& current_time)
{
    time = current_time;

    recordData();
    if (int(100 * time) % 2000 == 0)
        recordPose();
    pathFollowing();

    usv.simulate(eta_6, nu_6, n, n_c, psi_d, u_d, r_d, a_d);

    eta_3 = PoseSE2(eta_6(0), eta_6(1), eta_6(5));
    nu_3 = Eigen::Vector3d(nu_6(0), nu_6(1), nu_6(5));
}