#include "../include/Simulation.hpp"
#include <fstream>

void write_csv(std::string filename, std::vector<std::pair<std::string, std::vector<double>>> dataset)
{
    std::ofstream myFile(filename);
    for(int j = 0; j < dataset.size(); ++j)
    {
        myFile << dataset.at(j).first;
        if(j != dataset.size() - 1) myFile << ",";
    }
    myFile << "\n";
    
    // Send data to the stream
    for(int i = 0; i < dataset.at(0).second.size(); ++i)
    {
        for(int j = 0; j < dataset.size(); ++j)
        {
            myFile << dataset.at(j).second.at(i);
            if(j != dataset.size() - 1) myFile << ",";
        }
        myFile << "\n";
    }
    
    myFile.close();
}

void run_simulation(std::vector<Eigen::Vector2d>& map, std::vector<PoseSE2>& ASVs_eta0, std::vector<Eigen::Vector2d>& ASVs_goal, std::vector<double>& ASVs_nominal_speed, std::vector<bool>& activate_COLAV, std::vector<bool>& plot_opt, int& scenario)
{
    double sampleTime = 0.02;
    double simuTime = 150;
    int samples = simuTime/sampleTime;
    int n_ASV = ASVs_eta0.size();
    std::vector<Ship> ASVs;
    std::vector<bool> reached_goal(n_ASV,false);
    std::vector<bool> all_reached_goal(n_ASV,true);

    for (int i=0; i < n_ASV; ++i)
    {
        ShipConfig ASV_cfg;
        ASV_cfg.index = i;
        ASV_cfg.eta0 = ASVs_eta0[i];
        ASV_cfg.nu0 = Eigen::Vector3d(0,0,0);
        ASV_cfg.u_nominal = ASVs_nominal_speed[i];
        ASV_cfg.waypoints = {ASV_cfg.eta0.position(), ASVs_goal[i]};
        ASV_cfg.map = map;
        Ship ASV(ASV_cfg);
        ASVs.push_back(ASV);
    }

    int n_comb = combination(n_ASV);
    container1 dist_data(n_comb);
    data data_time;
    container3 ASVs_pose_data(n_ASV);
    container3 ASVs_dynamics_data(n_ASV);
    bool break_loop = false;
    
    // main loop
    for (int i = 0; i < samples; ++i)
    {
        double time = i * sampleTime;
        record(time,data_time);

        int idx1 = 0, idx2 = 1;
        for (int comb = 0; comb < n_comb; ++comb)
        {
            double distance = (ASVs[idx1].eta_3.position() - ASVs[idx2].eta_3.position()).norm();
            record(distance, dist_data[comb]);
            update_indices(idx1, idx2, n_ASV-1);
        }

        std::vector<std::pair<PoseSE2,Eigen::Vector3d>> ASV_motion_data;
        for (int asv = 0; asv < n_ASV; ++asv)
            ASV_motion_data.push_back(ASVs[asv].provideMotionData());
            
        for (int asv = 0; asv < n_ASV; ++asv)
        {
            if (activate_COLAV[asv])
                ASVs[asv].simulate(time,ASV_motion_data);
            else
                ASVs[asv].simulate(time);
        }

        for (int asv = 0; asv < n_ASV; ++asv)
        {
            if ( (ASVs[asv].eta_3.position() - ASVs[asv].wps.back()).norm() < 20 )
                reached_goal[asv] = true;
        }

        if (reached_goal == all_reached_goal)
            break;
    }

    // pack data
    for (int i = 0; i < n_ASV; ++i)
    {
        ASVs_pose_data[i] = packData(ASVs[i].data_x_, ASVs[i].data_y_, ASVs[i].data_psi_, ASVs[i].data_x, ASVs[i].data_y);
        ASVs_dynamics_data[i] = packData(ASVs[i].data_psi_ref, ASVs[i].data_psi_d, ASVs[i].data_psi, ASVs[i].data_u_ref, ASVs[i].data_u_d, ASVs[i].data_u);
    }

    // plot results
    if (plot_opt[2])
        plotDistance(data_time, dist_data, get_ASV_ASV(n_ASV), scenario, n_ASV, true);
    if (plot_opt[0])
        plotXY(ASVs_pose_data, scenario);
    if (plot_opt[1])
        plotDynamics(data_time, ASVs_dynamics_data, scenario, true);

    // std::vector<std::pair<std::string, std::vector<double>>> dataset;
    // dataset.push_back(std::make_pair("time",data_time));
    // for (int i=0; i<n_ASV;++i)
    // {
    //     dataset.push_back(std::make_pair("north",ASVs[i].data_x));
    //     dataset.push_back(std::make_pair("east",ASVs[i].data_y));
    //     dataset.push_back(std::make_pair("heading",ASVs[i].data_psi));
    // }

    // write_csv("scenario"+std::to_string(scenario)+".csv", dataset);
}

void run_simulation(std::vector<Ellipse> map_details, std::vector<PoseSE2>& ASVs_eta0, std::vector<std::vector<Eigen::Vector2d>>& waypoints, std::vector<double>& ASVs_nominal_speed)
{
    double sampleTime = 0.02;
    double simuTime = 200;
    int samples = simuTime/sampleTime;
    int n_ASV = ASVs_eta0.size();
    std::vector<Ship> ASVs;
    std::vector<bool> reached_goal(n_ASV,false);
    std::vector<bool> all_reached_goal(n_ASV,true);
    std::vector<Eigen::Vector2d> map;

    for (int i=0; i < map_details.size(); ++i)
    {
        auto e = map_details[i];
        auto obs = getEllipticObstacle(e.position,e.a,e.b,e.start,e.end,e.angle);
        map.insert(map.end(), obs.begin(), obs.end());
    }

    std::cout << "Map created! Size = " << map.size() << std::endl;

    for (int i=0; i < n_ASV; ++i)
    {
        ShipConfig ASV_cfg;
        ASV_cfg.index = i;
        ASV_cfg.eta0 = ASVs_eta0[i];
        ASV_cfg.nu0 = Eigen::Vector3d(0,0,0);
        ASV_cfg.u_nominal = ASVs_nominal_speed[i];
        ASV_cfg.waypoints = waypoints[i];
        ASV_cfg.map = map;
        Ship ASV(ASV_cfg);
        ASVs.push_back(ASV);
    }

    data data_time;
    container3 ASVs_pose_data(n_ASV);
    container3 ASVs_dynamics_data(n_ASV);
    container1 dist_data;

    int n_comb = 0;
    if (n_ASV>1)
    {
        n_comb = combination(n_ASV);
        dist_data.resize(n_comb);
    }

    bool break_loop = false;

    for (int i = 0; i < samples; ++i)
    {
        double time = i * sampleTime;
        record(time,data_time);

        int idx1 = 0, idx2 = 1;
        for (int comb = 0; comb < n_comb; ++comb)
        {
            double distance = (ASVs[idx1].eta_3.position() - ASVs[idx2].eta_3.position()).norm();
            record(distance, dist_data[comb]);
            update_indices(idx1, idx2, n_ASV-1);
        }

        std::vector<std::pair<PoseSE2,Eigen::Vector3d>> ASV_motion_data;
        for (int asv = 0; asv < n_ASV; ++asv)
            ASV_motion_data.push_back(ASVs[asv].provideMotionData());
            
        for (int asv = 0; asv < n_ASV; ++asv)
            ASVs[asv].simulate(time,ASV_motion_data);

        for (int asv = 0; asv < n_ASV; ++asv)
        {
            if ( (ASVs[asv].eta_3.position() - ASVs[asv].wps.back()).norm() < 20 )
                reached_goal[asv] = true;
        }

        if (reached_goal == all_reached_goal)
            break;
    }

    // pack data
    for (int i = 0; i < n_ASV; ++i)
    {
        ASVs_pose_data[i] = packData(ASVs[i].data_x_, ASVs[i].data_y_, ASVs[i].data_psi_, ASVs[i].data_x, ASVs[i].data_y);
        ASVs_dynamics_data[i] = packData(ASVs[i].data_psi_ref, ASVs[i].data_psi_d, ASVs[i].data_psi, ASVs[i].data_u_ref, ASVs[i].data_u_d, ASVs[i].data_u);
    }

    plotXY(ASVs_pose_data, map_details, waypoints);
    plotDynamics(data_time, ASVs_dynamics_data, 0, false);
    if (n_ASV > 1)
        {plotDistance(data_time, dist_data, get_ASV_ASV(n_ASV), 0, n_ASV, false);}
}


std::pair<PoseSE2, Eigen::Vector2d> coord(double radius, double angle)
{
    double x0 = radius * cos(D2R(angle));
    double y0 = radius * sin(D2R(angle));
    double psi0 = g2o::normalize_theta(D2R(angle-180));

    double xg = -x0;
    double yg = -y0;

    return {PoseSE2(x0, y0, psi0), Eigen::Vector2d(xg, yg)};
}