#include "../include/misc.hpp"

double D2R(double x){return x * M_PI/180;}
double R2D(double x){return x * 180/M_PI;}

Eigen::Matrix3d Smtrx(Eigen::Vector3d a)
{
    Eigen::Matrix3d S;
    S <<    0, -a(2),  a(1),
         a(2),     0, -a(0),
        -a(1),  a(0),     0;

    return S;
}

Eigen::MatrixXd Hmtrx(Eigen::Vector3d r)
{
    Eigen::MatrixXd H(6,6);
    H << Eigen::MatrixXd::Identity(6,6);
    H.block<3,3>(0,3) = Smtrx(r).transpose();

    return H;
}

Eigen::MatrixXd m2c(Eigen::MatrixXd& M, Eigen::VectorXd& nu)
{
    Eigen::MatrixXd M_ = 0.5 * (M + M.transpose());
    
    Eigen::Matrix3d M11 = M_.block<3,3>(0,0);
    Eigen::Matrix3d M12 = M_.block<3,3>(0,3);
    Eigen::Matrix3d M21 = M12.transpose();
    Eigen::Matrix3d M22 = M_.block<3,3>(3,3);

    Eigen::Vector3d nu1 = nu.head(3);
    Eigen::Vector3d nu2 = nu.tail<3>();

    Eigen::Vector3d dt_dnu1 = M11 * nu1 + M12 * nu2;
    Eigen::Vector3d dt_dnu2 = M21 * nu1 + M22 * nu2;

    Eigen::MatrixXd C = Eigen::MatrixXd::Zero(6,6);

    C.block<3,3>(0,3) = -Smtrx(dt_dnu1);
    C.block<3,3>(3,0) = -Smtrx(dt_dnu1);
    C.block<3,3>(3,3) = -Smtrx(dt_dnu2);
        
    return C;
}

double sat(double x, double x_min, double x_max)
{
    if (x > x_max)
        x = x_max;
    else if (x < x_min)
        x = x_min;
    
    return x;
}

Eigen::VectorXd crossFlowDrag(double L, double B, double T, Eigen::VectorXd nu_r)
{
    double rho = 1026;
    int n = 20;

    double dx = L/20;
    double Cd_2D = 0.99;

    double Yh = 0;
    double Nh = 0;
    double xL = -L/2;

    for (int i = 0; i < n+1; ++i)
    {
        double v_r = nu_r(1);
        double r = nu_r(5);
        double Ucf = fabs(v_r + xL * r) * (v_r + xL * r);
        Yh = Yh - 0.5 * rho * T * Cd_2D * Ucf * dx;
        Nh = Nh - 0.5 * rho * T * Cd_2D * xL * Ucf * dx;
        
        xL += dx;
    }
    Eigen::VectorXd tau_crossflow(6);
    tau_crossflow << 0, Yh, 0, 0, 0, Nh;

    return tau_crossflow;
}

Eigen::Matrix3d Rzyx(double phi, double theta, double psi)
{
    double cphi = cos(phi);
    double sphi = sin(phi);
    double cth = cos(theta);
    double sth = sin(theta);
    double cpsi = cos(psi);
    double spsi = sin(psi);

    Eigen::Matrix3d R;
    R << cpsi*cth, -spsi*cphi+cpsi*sth*sphi, spsi*sphi+cpsi*cphi*sth,
         spsi*cth,  cpsi*cphi+sphi*sth*spsi, -cpsi*sphi+sth*spsi*cphi,
         -sth,      cth*sphi,                cth*cphi;

    return R;
}

Eigen::Matrix3d Tzyx(double phi, double theta)
{
    double cphi = cos(phi);
    double sphi = sin(phi);
    double cth = cos(theta);
    double sth = sin(theta);

    Eigen::Matrix3d T;
    T << 1,  sphi*sth/cth,  cphi*sth/cth,
         0,  cphi,          -sphi,
         0,  sphi/cth,      cphi/cth;

    return T;
}

Eigen::Matrix3d R_psi(double psi)
{
    Eigen::Matrix3d R;
    R << cos(psi), -sin(psi), 0,
         sin(psi),  cos(psi), 0,
                0,         0, 1;
    return R;
}


void record(double data, std::vector<double>& container)
{
    container.push_back(data);
}

bool inInterval(double x, double lowerbound, double upperbound)
{
    if (x >= lowerbound && x <= upperbound)
        return true;
    else
        return false;
}

int fact(int n)
{
    if (n == 0||n==1)
    return 1;
    else
    return n*fact(n-1);
}

int combination(int n)
{
    return fact(n)/(fact(2)*fact(n-2));
}

void update_indices(int& idx1, int& idx2, int idx_max)
{
    if (idx2 < idx_max)
        idx2 += 1;
    else
    {
        idx1 += 1;
        idx2 = idx1 + 1;
    }            
}

std::vector<std::string> get_ASV_ASV(int n)
{
    int n_com = combination(n);
    std::vector<std::string> ASV_ASV(n_com);
    int x = 1, y = 2;

    for (int i = 0; i < n_com; ++i)
    {
        ASV_ASV[i] = "ASV " + std::to_string(x) + "-" + std::to_string(y);
        update_indices(x,y,n);
    }

    return ASV_ASV;
}

std::pair<double,int> getMinDistance(std::vector<Eigen::Vector2d> obstacles, Eigen::Vector2d position)
{
    std::vector<double> distances;

    for (int i = 0; i < obstacles.size(); i++)
    {
        double dist = (position - obstacles[i]).norm();
        distances.push_back(dist);
    }

    auto it = std::min_element(distances.begin(),distances.end());
    auto idx = int(std::distance(distances.begin(), it));
    double dist_min = *it;

    return {dist_min,idx};
}

std::vector<Eigen::Vector2d> getNearbyObstacles(Eigen::Vector2d& position, std::vector<Eigen::Vector2d>& static_obstacles, std::vector<Eigen::Vector2d>& dynamic_obstacles, double& range)
{
    std::vector<Eigen::Vector2d> obs;

    for (int i=0; i<static_obstacles.size(); ++i)
    {
        if ((position - static_obstacles[i]).norm() < range){
            obs.push_back(static_obstacles[i]);
        }
    }

    for (int i=0; i<dynamic_obstacles.size(); ++i)
    {
        if ((position - dynamic_obstacles[i]).norm() < range){
            obs.push_back(dynamic_obstacles[i]);
        }
    }

    return obs;
}


std::vector<Eigen::Vector2d> getEllipticObstacle(Eigen::Vector2d& position, double& a, double& b, double& start, double& end, double& angle)
{
    double d = 5;
    double step = d / std::max(a,b);
    double diff = D2R(end - start);
    int n = int(diff / step);

    std::vector<Eigen::Vector2d> obs;

    for (int i = 0; i < n; ++i)
    {
        double theta = D2R(start) + i * step,
        dx = a*cos(theta),
        dy = b*sin(theta),
        dx_rotated = dx*cos(D2R(angle)) - dy*sin(D2R(angle)),
        dy_rotated = dx*sin(D2R(angle)) + dy*cos(D2R(angle));
        auto obs_position = position + Eigen::Vector2d(dx_rotated,dy_rotated);
        obs.push_back(obs_position);
    }

    return obs;
}