#include "../include/Guidance.hpp"
#include "../include/USV.hpp"
#include <matplot/matplot.h>

using namespace matplot;

int main()
{
    double sampleTime = 0.1;
    double simuTime = 200;
    int samples = int(simuTime / sampleTime);

    std::vector<Eigen::Vector2d> waypoints;
    waypoints.push_back(Eigen::Vector2d(0,0));
    waypoints.push_back(Eigen::Vector2d(100,100));
    waypoints.push_back(Eigen::Vector2d(0,200));
    USV usv(0,0);
    Guidance guidance(waypoints);

    Eigen::VectorXd eta(6), nu(6);
    eta << 0, 0, 0, 0, 0, 0;
    nu << 0, 0, 0, 0, 0, 0;

    Eigen::Vector2d n, n_c;
    n << 0, 0;
    n_c << 0, 0;

    double psi_ref, psi_d, r_d, alpha_d;
    psi_ref = D2R(0);
    psi_d = eta(5);
    r_d = nu(5);
    alpha_d = 0;

    double u_ref, u_d, u_d_dot;
    u_ref = 2.5;
    u_d = nu(0);
    u_d_dot = 0;

    std::vector<double>
    data_time,
    data_x, data_y,
    data_psi_ref, data_psi_d, data_psi,
    data_u_ref, data_u_d, data_u,
    data_n1_c, data_n1,
    data_n2_c, data_n2;

    for (int i = 0; i < samples; ++i)
    {   
        double current_time = i * sampleTime;

        PoseSE2 eta_3DOF(eta(0),eta(1),eta(5));
        Eigen::Vector3d nu_3DOF(nu(0),nu(1),nu(5));

        record(current_time, data_time);
        record(eta_3DOF.x(), data_x);
        record(eta_3DOF.y(), data_y);
        record(R2D(eta_3DOF.theta()), data_psi);
        record(R2D(psi_ref), data_psi_ref);
        record(R2D(psi_d), data_psi_d);
        record(nu_3DOF(0),data_u);
        record(u_ref, data_u_ref);
        record(u_d, data_u_d);
        record(n_c(0), data_n1_c);
        record(n_c(1), data_n2_c);
        record(n(0), data_n1);
        record(n(1), data_n2);

        psi_ref = guidance.LOS(eta_3DOF);
        guidance.referenceModel(psi_ref, psi_d, r_d, alpha_d, u_ref, u_d, u_d_dot);
        usv.simulate(eta, nu, n, n_c, psi_d, u_d, r_d, u_d_dot);
        
        if ((eta_3DOF.position() - waypoints.back()).norm() < 1)
            break;
    }

    figure(1);
    plot(data_y, data_x);
    title("xy plot");
    xlabel("East (m)");
    ylabel("North (m)");
    grid(on);
    axis(automatic);
    show();

    // yaw & speed
    figure(2);
    subplot(2,1,0);
    title("Yaw angle");
    ylabel("degree");
    plot(data_time, data_psi_ref);
    hold(on);
    plot(data_time, data_psi_d);
    hold(on);
    plot(data_time, data_psi);
    grid(on);
    legend("reference","desired", "actual");

    subplot(2,1,1);
    title("Surge speed");
    xlabel("time (s)");
    ylabel("m/s");
    plot(data_time, data_u_ref);
    hold(on);
    plot(data_time, data_u_d);
    hold(on);
    plot(data_time, data_u);
    legend("reference","desired", "actual");
    grid(on);
    show();

    // propeller speed
    figure(3);
    subplot(2,1,0);
    title("Propeller speed 1");
    ylabel("rps");
    plot(data_time, data_n1_c);
    hold(on);
    plot(data_time, data_n1);
    grid(on);
    legend("commanded", "actual");

    subplot(2,1,1);
    title("Propeller speed 2");
    xlabel("time (s)");
    ylabel("rps");
    plot(data_time, data_n2_c);
    hold(on);
    plot(data_time, data_n2);
    grid(on);
    legend("commanded", "actual");
    show();
}